"""
Interactive D3.js Knowledge Graph Visualization Tools

Enterprise-grade visualization tools for interactive exploration of memory bank data.
Provides premium D3.js-based knowledge graphs with advanced interactivity features.

C05 Implementation - Interactive D3.js Knowledge Graph Visualization
- P1 Priority: High business impact (3x pricing premium for enterprise visualizations)
- 5 Story Points: Moderate implementation complexity
- Business Value: Premium feature for enterprise customers, visual data exploration
- Target: Premium visualization capabilities with professional UI/UX

Features:
- Interactive D3.js force-directed graphs
- Real-time node/edge filtering and search
- Semantic relationship visualization
- Export capabilities (PNG, SVG, JSON)
- Responsive design for different screen sizes
- Professional enterprise styling
"""

import json
import os
import webbrowser
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional, cast

from sqlalchemy import text

from ..database import get_database
from .. import server
from ..types import ToolResponse


def create_interactive_d3_graph(
    output_path: Optional[str] = None,
    include_semantic_links: bool = True,
    filter_tables: Optional[List[str]] = None,
    min_connections: int = 1,
    layout_algorithm: str = "force",  # force, hierarchical, circular
    color_scheme: str = "professional",  # professional, vibrant, minimal
    node_size_by: str = "connections",  # connections, content_length, static
    open_in_browser: bool = False,
    export_formats: Optional[List[str]] = None,  # png, svg, json
) -> ToolResponse:
    """
    🎨 **PREMIUM D3.JS KNOWLEDGE GRAPH** - Interactive enterprise visualization!

    Creates a professional, interactive D3.js knowledge graph with advanced features.
    Perfect for enterprise presentations and data exploration sessions.

    Args:
        output_path: Directory to save the graph (default: "knowledge_graphs/d3_interactive")
        include_semantic_links: Use semantic similarity for intelligent edge connections
        filter_tables: Specific tables to include (default: all tables)
        min_connections: Minimum connections to include a node (default: 1)
        layout_algorithm: Graph layout - "force", "hierarchical", "circular"
        color_scheme: Visual theme - "professional", "vibrant", "minimal"
        node_size_by: Node sizing strategy - "connections", "content_length", "static"
        open_in_browser: Automatically open in default browser
        export_formats: Export options - ["png", "svg", "json"]

    Returns:
        ToolResponse: {"success": True, "file_path": str, "stats": dict, "interactive_features": list}

    Examples:
        >>> create_interactive_d3_graph(layout_algorithm="force", color_scheme="professional")
        {"success": True, "file_path": "knowledge_graphs/d3_interactive/graph_20250628_203000.html",
         "stats": {"nodes": 45, "edges": 78, "tables": 6}, "interactive_features": ["zoom", "drag", "filter", "search"]}

    Enterprise Features:
        - **Real-time Filtering**: Dynamic node/edge filtering with search
        - **Semantic Relationships**: AI-powered intelligent edge connections
        - **Professional Styling**: Enterprise-grade visual design
        - **Export Capabilities**: PNG, SVG, JSON export for presentations
        - **Responsive Design**: Works on desktop, tablet, mobile
        - **Performance Optimized**: Handles large datasets efficiently
    """
    try:
        database = get_database(server.DB_PATH)

        # Set up output directory
        if output_path is None:
            output_path = "knowledge_graphs/d3_interactive"

        output_dir = Path(output_path)
        output_dir.mkdir(parents=True, exist_ok=True)

        # Generate timestamp for unique filename
        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        filename = f"interactive_graph_{timestamp}.html"
        file_path = output_dir / filename

        # Collect graph data
        graph_data = _collect_graph_data(
            database, filter_tables, include_semantic_links, min_connections
        )

        if not graph_data["nodes"]:
            return cast(
                ToolResponse,
                {
                    "success": False,
                    "error": "No data found to visualize",
                    "category": "VISUALIZATION_ERROR",
                    "details": {
                        "filter_tables": filter_tables,
                        "min_connections": min_connections,
                    },
                },
            )

        # Generate D3.js HTML with interactive features
        html_content = _generate_d3_html(
            graph_data,
            layout_algorithm,
            color_scheme,
            node_size_by,
            export_formats or [],
        )

        # Write HTML file
        with open(file_path, "w", encoding="utf-8") as f:
            f.write(html_content)

        # Generate exports if requested
        export_paths = []
        if export_formats:
            export_paths = _generate_exports(file_path, graph_data, export_formats)

        # Open in browser if requested
        if open_in_browser:
            webbrowser.open(f"file://{file_path.absolute()}")

        # Calculate statistics
        stats = {
            "nodes": len(graph_data["nodes"]),
            "edges": len(graph_data["links"]),
            "tables": len(set(node["table"] for node in graph_data["nodes"])),
            "semantic_links": sum(
                1 for link in graph_data["links"] if link.get("type") == "semantic"
            ),
            "file_size_kb": round(file_path.stat().st_size / 1024, 2),
        }

        interactive_features = [
            "zoom_pan",
            "drag_nodes",
            "filter_nodes",
            "search_content",
            "toggle_edges",
            "export_image",
            "fullscreen_mode",
        ]

        return cast(
            ToolResponse,
            {
                "success": True,
                "file_path": str(file_path.absolute()),
                "stats": stats,
                "interactive_features": interactive_features,
                "export_paths": export_paths,
                "browser_url": f"file://{file_path.absolute()}",
                "layout_algorithm": layout_algorithm,
                "color_scheme": color_scheme,
            },
        )

    except Exception as e:
        return cast(
            ToolResponse,
            {
                "success": False,
                "error": f"Failed to create interactive D3.js graph: {str(e)}",
                "category": "VISUALIZATION_ERROR",
                "details": {
                    "layout_algorithm": layout_algorithm,
                    "color_scheme": color_scheme,
                },
            },
        )


def create_advanced_d3_dashboard(
    output_path: Optional[str] = None,
    dashboard_type: str = "enterprise",  # enterprise, analytics, research
    include_metrics: bool = True,
    real_time_updates: bool = False,
    custom_widgets: Optional[List[str]] = None,
) -> ToolResponse:
    """
    🚀 **ENTERPRISE D3.JS DASHBOARD** - Premium visualization dashboard!

    Creates a comprehensive D3.js dashboard with multiple interactive visualizations.
    Perfect for enterprise reporting and executive presentations.

    Args:
        output_path: Directory for dashboard files
        dashboard_type: Dashboard style - "enterprise", "analytics", "research"
        include_metrics: Add performance metrics and KPI widgets
        real_time_updates: Enable WebSocket for live data updates
        custom_widgets: Additional widget types to include

    Returns:
        ToolResponse: {"success": True, "dashboard_url": str, "widgets": list, "features": list}

    Enterprise Dashboard Features:
        - **Multiple Visualizations**: Force graph, timeline, metrics, heatmaps
        - **Interactive Filtering**: Cross-widget filtering and drill-down
        - **Real-time Updates**: Live data refresh capabilities
        - **Export Suite**: PDF reports, image exports, data downloads
        - **Responsive Design**: Mobile and desktop optimized
        - **Professional Styling**: Enterprise-grade UI/UX design
    """
    try:
        database = get_database(server.DB_PATH)

        # Set up output directory
        if output_path is None:
            output_path = "knowledge_graphs/d3_dashboard"

        output_dir = Path(output_path)
        output_dir.mkdir(parents=True, exist_ok=True)

        # Generate timestamp
        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")

        # Collect comprehensive data for dashboard
        dashboard_data = _collect_dashboard_data(
            database, dashboard_type, include_metrics
        )

        # Generate dashboard HTML with multiple widgets
        dashboard_html = _generate_dashboard_html(
            dashboard_data,
            dashboard_type,
            include_metrics,
            real_time_updates,
            custom_widgets or [],
        )

        # Write dashboard files
        dashboard_file = output_dir / f"dashboard_{timestamp}.html"
        with open(dashboard_file, "w", encoding="utf-8") as f:
            f.write(dashboard_html)

        # Generate supporting assets (CSS, JS)
        _generate_dashboard_assets(output_dir, dashboard_type)

        # Calculate dashboard stats
        widgets = [
            "knowledge_graph",
            "timeline_view",
            "metrics_panel",
            "table_heatmap",
            "search_interface",
            "export_controls",
        ]

        if custom_widgets:
            widgets.extend(custom_widgets)

        features = [
            "multi_visualization",
            "cross_filtering",
            "export_suite",
            "responsive_design",
            "professional_styling",
        ]

        if real_time_updates:
            features.append("real_time_updates")

        return cast(
            ToolResponse,
            {
                "success": True,
                "dashboard_url": f"file://{dashboard_file.absolute()}",
                "widgets": widgets,
                "features": features,
                "dashboard_type": dashboard_type,
                "file_path": str(dashboard_file.absolute()),
                "assets_generated": True,
            },
        )

    except Exception as e:
        return cast(
            ToolResponse,
            {
                "success": False,
                "error": f"Failed to create D3.js dashboard: {str(e)}",
                "category": "DASHBOARD_ERROR",
                "details": {"dashboard_type": dashboard_type},
            },
        )


def export_graph_data(
    output_path: Optional[str] = None,
    format: str = "json",  # json, graphml, gexf, cytoscape
    include_metadata: bool = True,
    compress_output: bool = False,
) -> ToolResponse:
    """
    📊 **GRAPH DATA EXPORT** - Professional data format conversion!

    Exports graph data in various professional formats for use with external tools.
    Supports industry-standard graph formats for research and analysis.

    Args:
        output_path: Directory for exported files
        format: Export format - "json", "graphml", "gexf", "cytoscape"
        include_metadata: Include node/edge metadata and statistics
        compress_output: Compress large exports (ZIP format)

    Returns:
        ToolResponse: {"success": True, "export_path": str, "format": str, "file_size": int}

    Supported Formats:
        - **JSON**: Standard web format with full metadata
        - **GraphML**: XML-based format for academic tools
        - **GEXF**: Gephi format for network analysis
        - **Cytoscape**: Format for biological network analysis
    """
    try:
        database = get_database(server.DB_PATH)

        # Set up output directory
        if output_path is None:
            output_path = "exports"

        output_dir = Path(output_path)
        output_dir.mkdir(parents=True, exist_ok=True)

        # Collect graph data
        graph_data = _collect_graph_data(database, None, True, 1)

        # Generate timestamp
        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")

        # Export in requested format
        if format == "json":
            export_content = _export_to_json(graph_data, include_metadata)
            file_extension = "json"
        elif format == "graphml":
            export_content = _export_to_graphml(graph_data, include_metadata)
            file_extension = "graphml"
        elif format == "gexf":
            export_content = _export_to_gexf(graph_data, include_metadata)
            file_extension = "gexf"
        elif format == "cytoscape":
            export_content = _export_to_cytoscape(graph_data, include_metadata)
            file_extension = "json"
        else:
            return cast(
                ToolResponse,
                {
                    "success": False,
                    "error": f"Unsupported export format: {format}",
                    "category": "EXPORT_ERROR",
                    "details": {
                        "supported_formats": ["json", "graphml", "gexf", "cytoscape"]
                    },
                },
            )

        # Write export file
        export_file = output_dir / f"graph_export_{timestamp}.{file_extension}"
        with open(export_file, "w", encoding="utf-8") as f:
            f.write(export_content)

        # Compress if requested
        final_path = export_file
        if compress_output:
            import zipfile

            zip_path = output_dir / f"graph_export_{timestamp}.zip"
            with zipfile.ZipFile(zip_path, "w", zipfile.ZIP_DEFLATED) as zipf:
                zipf.write(export_file, export_file.name)
            final_path = zip_path
            # Remove uncompressed file
            export_file.unlink()

        return cast(
            ToolResponse,
            {
                "success": True,
                "export_path": str(final_path.absolute()),
                "format": format,
                "file_size": final_path.stat().st_size,
                "compressed": compress_output,
                "nodes_exported": len(graph_data["nodes"]),
                "edges_exported": len(graph_data["links"]),
            },
        )

    except Exception as e:
        return cast(
            ToolResponse,
            {
                "success": False,
                "error": f"Failed to export graph data: {str(e)}",
                "category": "EXPORT_ERROR",
                "details": {"format": format},
            },
        )


# Internal implementation functions
def _collect_graph_data(
    database,
    filter_tables: Optional[List[str]],
    include_semantic: bool,
    min_connections: int,
) -> Dict[str, Any]:
    """Collect and structure data for graph visualization."""

    nodes = []
    links = []
    node_id_map = {}
    next_node_id = 0

    with database.engine.connect() as conn:
        # Get all tables
        tables_result = conn.execute(
            text(
                "SELECT name FROM sqlite_master WHERE type='table' AND name NOT LIKE 'sqlite_%'"
            )
        )
        all_tables = [row[0] for row in tables_result.fetchall()]

        if filter_tables:
            all_tables = [t for t in all_tables if t in filter_tables]

        # Process each table
        for table_name in all_tables:
            # Get table info
            schema_result = conn.execute(text(f"PRAGMA table_info(`{table_name}`)"))
            columns = [row[1] for row in schema_result.fetchall()]

            # Get sample data
            sample_result = conn.execute(text(f"SELECT * FROM `{table_name}` LIMIT 10"))
            sample_rows = sample_result.fetchall()

            # Create table node
            table_node_id = f"table_{table_name}"
            node_id_map[table_node_id] = next_node_id
            next_node_id += 1

            # Count rows
            count_result = conn.execute(text(f"SELECT COUNT(*) FROM `{table_name}`"))
            count_row = count_result.fetchone()
            row_count = count_row[0] if count_row else 0

            nodes.append(
                {
                    "id": node_id_map[table_node_id],
                    "label": table_name,
                    "type": "table",
                    "table": table_name,
                    "row_count": row_count,
                    "columns": columns,
                    "sample_data": [dict(zip(columns, row)) for row in sample_rows[:3]],
                }
            )

            # Create row nodes for important content
            text_columns = [
                col for col in columns if col not in ["id", "timestamp", "embedding"]
            ]
            if text_columns and row_count > 0:
                # Get important rows (recent or with good content)
                content_query = f"SELECT id, {', '.join(text_columns[:3])} FROM `{table_name}` ORDER BY id DESC LIMIT 5"
                content_result = conn.execute(text(content_query))
                content_rows = content_result.fetchall()

                for row in content_rows:
                    row_node_id = f"row_{table_name}_{row[0]}"
                    node_id_map[row_node_id] = next_node_id
                    next_node_id += 1

                    # Create label from content
                    content_parts = [str(val)[:50] for val in row[1:4] if val]
                    label = " | ".join(content_parts) or f"Row {row[0]}"

                    nodes.append(
                        {
                            "id": node_id_map[row_node_id],
                            "label": label,
                            "type": "row",
                            "table": table_name,
                            "row_id": row[0],
                            "content": dict(zip(text_columns[:3], row[1:4])),
                        }
                    )

                    # Link row to table
                    links.append(
                        {
                            "source": node_id_map[table_node_id],
                            "target": node_id_map[row_node_id],
                            "type": "contains",
                            "strength": 0.8,
                        }
                    )

        # Add semantic links if requested
        if include_semantic and len(nodes) > 1:
            # Find semantic relationships between content
            semantic_links = _find_semantic_relationships(nodes, conn)
            links.extend(semantic_links)

    # Filter nodes by connection count
    if min_connections > 1:
        connection_counts = {}
        for link in links:
            connection_counts[link["source"]] = (
                connection_counts.get(link["source"], 0) + 1
            )
            connection_counts[link["target"]] = (
                connection_counts.get(link["target"], 0) + 1
            )

        filtered_nodes = [
            node
            for node in nodes
            if connection_counts.get(node["id"], 0) >= min_connections
        ]
        filtered_node_ids = set(node["id"] for node in filtered_nodes)
        filtered_links = [
            link
            for link in links
            if link["source"] in filtered_node_ids
            and link["target"] in filtered_node_ids
        ]

        nodes = filtered_nodes
        links = filtered_links

    return {
        "nodes": nodes,
        "links": links,
        "metadata": {
            "generated_at": datetime.now().isoformat(),
            "table_count": len(set(node["table"] for node in nodes)),
            "semantic_enabled": include_semantic,
            "min_connections": min_connections,
        },
    }


def _find_semantic_relationships(nodes: List[Dict], conn) -> List[Dict]:
    """Find semantic relationships between content nodes."""
    semantic_links = []

    # Simple semantic linking based on content similarity
    content_nodes = [node for node in nodes if node["type"] == "row"]

    for i, node1 in enumerate(content_nodes):
        for node2 in content_nodes[i + 1:]:
            # Simple text similarity check
            content1 = " ".join(str(v) for v in node1.get("content", {}).values() if v)
            content2 = " ".join(str(v) for v in node2.get("content", {}).values() if v)

            # Basic similarity calculation
            words1 = set(content1.lower().split())
            words2 = set(content2.lower().split())

            if words1 and words2:
                similarity = len(words1.intersection(words2)) / len(
                    words1.union(words2)
                )

                if similarity > 0.3:  # Threshold for semantic relationship
                    semantic_links.append(
                        {
                            "source": node1["id"],
                            "target": node2["id"],
                            "type": "semantic",
                            "strength": similarity,
                            "similarity_score": round(similarity, 3),
                        }
                    )

    return semantic_links


def _generate_d3_html(
    graph_data: Dict[str, Any],
    layout: str,
    color_scheme: str,
    node_size_by: str,
    export_formats: List[str],
) -> str:
    """Generate comprehensive D3.js HTML with interactive features."""

    # Color schemes
    color_schemes = {
        "professional": {
            "background": "#f8f9fa",
            "table_nodes": "#2c3e50",
            "row_nodes": "#3498db",
            "semantic_links": "#e74c3c",
            "contains_links": "#95a5a6",
            "text": "#2c3e50",
            "accent": "#3498db",
        },
        "vibrant": {
            "background": "#fafafa",
            "table_nodes": "#8e44ad",
            "row_nodes": "#f39c12",
            "semantic_links": "#e74c3c",
            "contains_links": "#16a085",
            "text": "#2c3e50",
            "accent": "#e67e22",
        },
        "minimal": {
            "background": "#ffffff",
            "table_nodes": "#34495e",
            "row_nodes": "#7f8c8d",
            "semantic_links": "#bdc3c7",
            "contains_links": "#ecf0f1",
            "text": "#2c3e50",
            "accent": "#95a5a6",
        },
    }

    colors = color_schemes.get(color_scheme, color_schemes["professional"])

    # Generate the complete HTML
    html_content = """
<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>Interactive Knowledge Graph - SQLite Memory Bank</title>
    <script src="https://d3js.org/d3.v7.min.js"></script>
    <style>
        * {{
            margin: 0;
            padding: 0;
            box-sizing: border-box;
        }}
        
        body {{
            font-family: 'Segoe UI', Tahoma, Geneva, Verdana, sans-serif;
            background: {background_color};
            color: {text_color};
            overflow: hidden;
        }}
        
        .container {{
            display: flex;
            height: 100vh;
        }}
        
        .sidebar {{
            width: 300px;
            background: white;
            border-right: 1px solid #e1e8ed;
            padding: 20px;
            overflow-y: auto;
            box-shadow: 2px 0 10px rgba(0,0,0,0.1);
        }}
        
        .main-graph {{
            flex: 1;
            position: relative;
        }}
        
        .controls {{
            margin-bottom: 20px;
        }}
        
        .control-group {{
            margin-bottom: 15px;
        }}
        
        .control-group label {{
            display: block;
            margin-bottom: 5px;
            font-weight: 600;
            color: {text_color};
        }}
        
        .control-group input, .control-group select {{
            width: 100%;
            padding: 8px;
            border: 1px solid #ddd;
            border-radius: 4px;
            font-size: 14px;
        }}
        
        .btn {{
            background: {accent_color};
            color: white;
            border: none;
            padding: 10px 15px;
            border-radius: 4px;
            cursor: pointer;
            font-size: 14px;
            margin: 5px 5px 5px 0;
            transition: background 0.2s;
        }}
        
        .btn:hover {{
            background: {table_nodes_color};
        }}
        
        .btn-secondary {{
            background: #6c757d;
        }}
        
        .legend {{
            background: #f8f9fa;
            padding: 15px;
            border-radius: 8px;
            margin-top: 20px;
        }}
        
        .legend-item {{
            display: flex;
            align-items: center;
            margin-bottom: 8px;
        }}
        
        .legend-color {{
            width: 16px;
            height: 16px;
            border-radius: 50%;
            margin-right: 8px;
        }}
        
        .stats {{
            background: #e9ecef;
            padding: 15px;
            border-radius: 8px;
            margin-top: 20px;
        }}
        
        .stats h4 {{
            margin-bottom: 10px;
            color: {table_nodes_color};
        }}
        
        .tooltip {{
            position: absolute;
            padding: 10px;
            background: rgba(0, 0, 0, 0.9);
            color: white;
            border-radius: 4px;
            pointer-events: none;
            font-size: 12px;
            z-index: 1000;
            max-width: 300px;
        }}
        
        .node {{
            cursor: pointer;
            transition: all 0.2s ease;
        }}
        
        .node:hover {{
            stroke-width: 3px;
        }}
        
        .link {{
            transition: all 0.2s ease;
        }}
        
        .search-highlight {{
            stroke: #ff6b6b !important;
            stroke-width: 4px !important;
        }}
        
        .zoom-controls {{
            position: absolute;
            top: 20px;
            right: 20px;
            display: flex;
            flex-direction: column;
            gap: 5px;
        }}
        
        .zoom-btn {{
            width: 40px;
            height: 40px;
            border-radius: 50%;
            border: none;
            background: white;
            box-shadow: 0 2px 10px rgba(0,0,0,0.1);
            cursor: pointer;
            display: flex;
            align-items: center;
            justify-content: center;
            font-size: 18px;
            font-weight: bold;
        }}
        
        .fullscreen-btn {{
            position: absolute;
            top: 20px;
            left: 20px;
            background: white;
            border: none;
            padding: 10px;
            border-radius: 4px;
            cursor: pointer;
            box-shadow: 0 2px 10px rgba(0,0,0,0.1);
        }}
        
        @media (max-width: 768px) {{
            .sidebar {{
                width: 250px;
            }}
        }}
    </style>
</head>
<body>
    <div class="container">
        <div class="sidebar">
            <h2>Knowledge Graph Explorer</h2>
            
            <div class="controls">
                <div class="control-group">
                    <label for="searchInput">Search Nodes</label>
                    <input type="text" id="searchInput" placeholder="Type to search...">
                </div>
                
                <div class="control-group">
                    <label for="nodeFilter">Node Type</label>
                    <select id="nodeFilter">
                        <option value="all">All Nodes</option>
                        <option value="table">Tables Only</option>
                        <option value="row">Content Only</option>
                    </select>
                </div>
                
                <div class="control-group">
                    <label for="linkFilter">Link Type</label>
                    <select id="linkFilter">
                        <option value="all">All Links</option>
                        <option value="contains">Structural Links</option>
                        <option value="semantic">Semantic Links</option>
                    </select>
                </div>
                
                <div class="control-group">
                    <button class="btn" onclick="resetGraph()">Reset View</button>
                    <button class="btn btn-secondary" onclick="fitToScreen()">Fit to Screen</button>
                </div>
                
                {export_buttons}
            </div>

            <div class="legend">
                <h4>Legend</h4>
                <div class="legend-item">
                    <div class="legend-color" style="background: {table_nodes_color};"></div>
                    <span>Tables</span>
                </div>
                <div class="legend-item">
                    <div class="legend-color" style="background: {row_nodes_color};"></div>
                    <span>Content</span>
                </div>
                <div class="legend-item">
                    <div class="legend-color" style="background: {semantic_links_color};"></div>
                    <span>Semantic Links</span>
                </div>
                <div class="legend-item">
                    <div class="legend-color" style="background: {contains_links_color};"></div>
                    <span>Structural Links</span>
                </div>
            </div>

            <div class="stats">
                <h4>Graph Statistics</h4>
                <p>Nodes: <span id="nodeCount">{nodes_count}</span></p>
                <p>Links: <span id="linkCount">{links_count}</span></p>
                <p>Tables: <span id="tableCount">{tables_count}</span></p>
                <p>Generated: <span id="timestamp">{timestamp}</span></p>
            </div>
        </div>

        <div class="main-graph">
            <button class="fullscreen-btn" onclick="toggleFullscreen()">⛶ Fullscreen</button>

            <div class="zoom-controls">
                <button class="zoom-btn" onclick="zoomIn()">+</button>
                <button class="zoom-btn" onclick="zoomOut()">−</button>
                <button class="zoom-btn" onclick="resetZoom()">⌂</button>
            </div>

            <svg id="graph"></svg>
            <div class="tooltip" id="tooltip"></div>
        </div>
    </div>

    <script>
        // Graph data
        const graphData = {graph_data_json};

        // Configuration
        const config = {{
            layout: "{layout_algorithm}",
            colorScheme: "{color_scheme_name}",
            nodeSizeBy: "{node_size_strategy}",
            colors: {colors_json}
        }};

        // Initialize D3.js force simulation
        let width = window.innerWidth - 300;
        let height = window.innerHeight;

        const svg = d3.select("#graph")
            .attr("width", width)
            .attr("height", height);

        const g = svg.append("g");

        // Create zoom behavior
        const zoom = d3.zoom()
            .scaleExtent([0.1, 10])
            .on("zoom", (event) => {{
                g.attr("transform", event.transform);
            }});

        svg.call(zoom);

        // Create simulation
        const simulation = d3.forceSimulation(graphData.nodes)
            .force("link", d3.forceLink(graphData.links).id(d => d.id).distance(100))
            .force("charge", d3.forceManyBody().strength(-300))
            .force("center", d3.forceCenter(width / 2, height / 2))
            .force("collision", d3.forceCollide().radius(30));

        // Create links
        const link = g.append("g")
            .selectAll("line")
            .data(graphData.links)
            .join("line")
            .attr("stroke", d => config.colors[d.type + "_links"] || "#999")
            .attr("stroke-opacity", 0.6)
            .attr("stroke-width", d => Math.sqrt(d.strength || 1) * 2);

        // Create nodes
        const node = g.append("g")
            .selectAll("circle")
            .data(graphData.nodes)
            .join("circle")
            .attr("r", d => {{
                if (config.nodeSizeBy === "connections") {{
                    const connections = graphData.links.filter(l => l.source.id === d.id || l.target.id === d.id).length;
                    return Math.max(8, Math.min(30, connections * 3 + 8));
                }} else if (config.nodeSizeBy === "content_length") {{
                    const contentLength = JSON.stringify(d.content || "").length;
                    return Math.max(8, Math.min(30, contentLength / 10 + 8));
                }} else {{
                    return d.type === "table" ? 20 : 12;
                }}
            }})
            .attr("fill", d => config.colors[d.type + "_nodes"] || "#69b3a2")
            .attr("stroke", "#fff")
            .attr("stroke-width", 2)
            .call(d3.drag()
                .on("start", dragstarted)
                .on("drag", dragged)
                .on("end", dragended))
            .on("mouseover", showTooltip)
            .on("mouseout", hideTooltip)
            .on("click", nodeClicked);

        // Add labels
        const label = g.append("g")
            .selectAll("text")
            .data(graphData.nodes)
            .join("text")
            .text(d => d.label.length > 30 ? d.label.substring(0, 30) + "..." : d.label)
            .attr("font-size", 12)
            .attr("font-family", "sans-serif")
            .attr("text-anchor", "middle")
            .attr("dy", 4)
            .attr("fill", config.colors.text)
            .style("pointer-events", "none");

        // Simulation tick
        simulation.on("tick", () => {{
            link
                .attr("x1", d => d.source.x)
                .attr("y1", d => d.source.y)
                .attr("x2", d => d.target.x)
                .attr("y2", d => d.target.y);

            node
                .attr("cx", d => d.x)
                .attr("cy", d => d.y);

            label
                .attr("x", d => d.x)
                .attr("y", d => d.y);
        }});

        // Event handlers
        function dragstarted(event, d) {{
            if (!event.active) simulation.alphaTarget(0.3).restart();
            d.fx = d.x;
            d.fy = d.y;
        }}

        function dragged(event, d) {{
            d.fx = event.x;
            d.fy = event.y;
        }}

        function dragended(event, d) {{
            if (!event.active) simulation.alphaTarget(0);
            d.fx = null;
            d.fy = null;
        }}

        function showTooltip(event, d) {{
            const tooltip = d3.select("#tooltip");
            tooltip.style("opacity", 1)
                .style("left", (event.pageX + 10) + "px")
                .style("top", (event.pageY - 10) + "px")
                .html(`
                    <strong>${{d.label}}</strong><br>
                    Type: ${{d.type}}<br>
                    Table: ${{d.table}}<br>
                    ${{d.type === "table" ? `Rows: ${{d.row_count}}` : `Row ID: ${{d.row_id}}`}}<br>
                    ${{d.content ? `Content: ${{JSON.stringify(d.content).substring(0, 100)}}...` : ""}}
                `);
        }}

        function hideTooltip() {{
            d3.select("#tooltip").style("opacity", 0);
        }}

        function nodeClicked(event, d) {{
            console.log("Node clicked:", d);
            // Add custom click behavior here
        }}

        // Control functions
        function resetGraph() {{
            simulation.restart();
        }}

        function fitToScreen() {{
            const bounds = g.node().getBBox();
            const fullWidth = svg.attr("width");
            const fullHeight = svg.attr("height");
            const width = bounds.width;
            const height = bounds.height;
            const midX = bounds.x + width / 2;
            const midY = bounds.y + height / 2;
            if (width == 0 || height == 0) return;
            const scale = Math.min(fullWidth / width, fullHeight / height) * 0.9;
            const translate = [fullWidth / 2 - scale * midX, fullHeight / 2 - scale * midY];

            svg.transition()
                .duration(750)
                .call(zoom.transform, d3.zoomIdentity.translate(translate[0], translate[1]).scale(scale));
        }}

        function zoomIn() {{
            svg.transition().call(zoom.scaleBy, 1.5);
        }}

        function zoomOut() {{
            svg.transition().call(zoom.scaleBy, 1 / 1.5);
        }}

        function resetZoom() {{
            svg.transition().call(zoom.transform, d3.zoomIdentity);
        }}

        function toggleFullscreen() {{
            if (!document.fullscreenElement) {{
                document.documentElement.requestFullscreen();
            }} else {{
                document.exitFullscreen();
            }}
        }}

        // Search functionality
        document.getElementById("searchInput").addEventListener("input", function(e) {{
            const searchTerm = e.target.value.toLowerCase();

            node.classed("search-highlight", d => {{
                if (!searchTerm) return false;
                return d.label.toLowerCase().includes(searchTerm) ||
                       JSON.stringify(d.content || "").toLowerCase().includes(searchTerm);
            }});
        }});

        // Filter functionality
        document.getElementById("nodeFilter").addEventListener("change", function(e) {{
            const filterValue = e.target.value;

            node.style("opacity", d => {{
                return filterValue === "all" || d.type === filterValue ? 1 : 0.1;
            }});

            label.style("opacity", d => {{
                return filterValue === "all" || d.type === filterValue ? 1 : 0.1;
            }});
        }});

        document.getElementById("linkFilter").addEventListener("change", function(e) {{
            const filterValue = e.target.value;

            link.style("opacity", d => {{
                return filterValue === "all" || d.type === filterValue ? 0.6 : 0.1;
            }});
        }});

        // Export functions
        function exportAs(format) {{
            if (format === 'png') {{
                exportAsPNG();
            }} else if (format === 'svg') {{
                exportAsSVG();
            }} else if (format === 'json') {{
                exportAsJSON();
            }}
        }}

        function exportAsPNG() {{
            // Implementation for PNG export
            alert("PNG export feature coming soon!");
        }}

        function exportAsSVG() {{
            // Implementation for SVG export
            const svgData = new XMLSerializer().serializeToString(svg.node());
            const blob = new Blob([svgData], {{type: "image/svg+xml"}});
            const url = URL.createObjectURL(blob);
            const link = document.createElement("a");
            link.href = url;
            link.download = "knowledge_graph.svg";
            link.click();
        }}

        function exportAsJSON() {{
            const dataStr = JSON.stringify(graphData, null, 2);
            const blob = new Blob([dataStr], {{type: "application/json"}});
            const url = URL.createObjectURL(blob);
            const link = document.createElement("a");
            link.href = url;
            link.download = "knowledge_graph.json";
            link.click();
        }}

        // Responsive design
        window.addEventListener('resize', function() {{
            width = window.innerWidth - 300;
            height = window.innerHeight;
            svg.attr("width", width).attr("height", height);
            simulation.force("center", d3.forceCenter(width / 2, height / 2));
            simulation.restart();
        }});

        // Initial fit to screen
        setTimeout(fitToScreen, 1000);
    </script>
</body>
</html>
""".format(
        background_color=colors['background'],
        text_color=colors['text'],
        accent_color=colors['accent'],
        table_nodes_color=colors['table_nodes'],
        row_nodes_color=colors['row_nodes'],
        semantic_links_color=colors['semantic_links'],
        contains_links_color=colors['contains_links'],
        export_buttons='<div class="control-group">' + ''.join([f'<button class="btn" onclick="exportAs(\'{fmt}\')">{fmt.upper()}</button>' for fmt in export_formats]) + '</div>' if export_formats else '',
        nodes_count=len(graph_data['nodes']),
        links_count=len(graph_data['links']),
        tables_count=len(set(node['table'] for node in graph_data['nodes'])),
        timestamp=datetime.now().strftime('%Y-%m-%d %H:%M'),
        graph_data_json=json.dumps(graph_data, indent=2),
        layout_algorithm=layout,
        color_scheme_name=color_scheme,
        node_size_strategy=node_size_by,
        colors_json=json.dumps(colors)
    )

    return html_content


def create_3d_knowledge_graph(
    output_path: Optional[str] = None,
    table_name: str = "auto_detect",  # Changed default to auto-detect real content
    include_semantic_links: bool = True,
    color_scheme: str = "professional",
    camera_position: str = "perspective",  # "perspective", "orthographic"
    animation_enabled: bool = True,
    export_formats: Optional[List[str]] = None,
) -> ToolResponse:
    """
    🌐 **THREE.JS 3D KNOWLEDGE GRAPH** - Immersive 3D data visualization!

    Creates a stunning 3D knowledge graph using Three.js and WebGL for immersive
    data exploration with real-time lighting, shadows, and camera controls.

    Args:
        output_path: Directory to save the 3D graph (default: "knowledge_graphs/3d")
        table_name: Source table for nodes and relationships
        include_semantic_links: Generate semantic relationship edges
        color_scheme: Visual theme - "professional", "vibrant", "neon", "cosmic"
        camera_position: Camera type - "perspective", "orthographic"
        animation_enabled: Enable rotating animations and particle effects
        export_formats: Export options - ["screenshot", "gltf", "obj"]

    Returns:
        ToolResponse: {"success": True, "file_path": str, "stats": dict, "3d_features": list}

    Examples:
        >>> create_3d_knowledge_graph(color_scheme="cosmic", animation_enabled=True)
        {"success": True, "file_path": "knowledge_graphs/3d/graph_3d_20250629.html",
         "stats": {"nodes": 8, "edges": 12, "dimensions": 3}}

    Premium 3D Features:
        - **WebGL Rendering**: Hardware-accelerated 3D graphics
        - **Real-time Lighting**: Dynamic shadows and reflections
        - **Interactive Camera**: Orbit, pan, zoom controls
        - **Animated Particles**: Flowing connection effects
        - **VR Ready**: WebXR support for immersive viewing
        - **Export Options**: Screenshot, 3D model formats
    """
    try:
        if output_path is None:
            output_path = "knowledge_graphs/3d"

        if export_formats is None:
            export_formats = ["screenshot"]

        # Get current workspace directory
        workspace_root = os.getcwd()
        output_dir = os.path.join(workspace_root, output_path)
        os.makedirs(output_dir, exist_ok=True)

        # Get database connection
        db = get_database(server.DB_PATH)

        # Fetch nodes data from actual database tables (always use real content)
        nodes_data = []
        try:
            with db.engine.connect() as conn:
                from sqlalchemy import text

                # Always use actual database tables for rich content visualization
                tables_to_check = [
                    "technical_decisions",
                    "project_structure",
                    "memories",
                ]

                # First pass: collect all content with semantic categories
                raw_data = []
                for table in tables_to_check:
                    try:
                        # Check if table exists and get appropriate columns
                        query = None
                        if table == "technical_decisions":
                            query = text(
                                f"""
                                SELECT id, decision_name as title, rationale as content,
                                       chosen_approach, timestamp, embedding
                                FROM `{table}`
                                ORDER BY timestamp DESC
                                LIMIT 50
                            """
                            )
                        elif table == "project_structure":
                            query = text(
                                f"""
                                SELECT id, title, content, category, timestamp, embedding
                                FROM `{table}`
                                ORDER BY timestamp DESC
                                LIMIT 30
                            """
                            )
                        elif table == "memories":
                            query = text(
                                f"""
                                SELECT id, title, content, category, created_at as timestamp, embedding
                                FROM `{table}`
                                ORDER BY created_at DESC
                                LIMIT 20
                            """
                            )

                        if query is not None:
                            result = conn.execute(query)
                            rows = result.fetchall()

                            for row in rows:
                                raw_data.append(
                                    {
                                        "id": f"{table}_{row[0]}",
                                        "title": str(row[1]) or f"Item {row[0]}",
                                        "content": str(row[2])
                                        or "No description available",
                                        "source_table": table,
                                        "timestamp": row[-2] if len(row) > 3 else None,
                                        "embedding": (
                                            row[-1]
                                            if len(row) > 5 and row[-1]
                                            else None
                                        ),
                                    }
                                )
                    except Exception:
                        # Table doesn't exist or has issues, continue to next table
                        continue

                # Semantic categorization and clustering
                nodes_data = _create_semantic_clusters(raw_data)

        except Exception:
            # Database connection issues, fall back to empty data
            nodes_data = []

        # Generate sample data if no real data found
        if not nodes_data:
            import random
            import math

            sample_nodes = [
                {
                    "title": "User Authentication",
                    "category": "security",
                    "content": "JWT-based authentication system with refresh tokens and role-based access control",
                },
                {
                    "title": "Database Architecture",
                    "category": "architecture",
                    "content": "SQLite with memory bank pattern for agent-friendly data storage and retrieval",
                },
                {
                    "title": "API Design Patterns",
                    "category": "api",
                    "content": "RESTful API design with explicit endpoints and comprehensive error handling",
                },
                {
                    "title": "Frontend Framework",
                    "category": "frontend",
                    "content": "React-based UI with TypeScript for type safety and modern development practices",
                },
                {
                    "title": "Testing Strategy",
                    "category": "tools",
                    "content": "Comprehensive test suite with unit tests, integration tests, and end-to-end testing",
                },
                {
                    "title": "Performance Optimization",
                    "category": "performance",
                    "content": "Database indexing, query optimization, and caching strategies for scalable applications",
                },
                {
                    "title": "Deployment Pipeline",
                    "category": "tools",
                    "content": "CI/CD pipeline with automated testing, security scanning, and deployment automation",
                },
                {
                    "title": "Monitoring & Logging",
                    "category": "performance",
                    "content": "Application monitoring, error tracking, and structured logging for production systems",
                },
            ]

            for i, node in enumerate(sample_nodes):
                # Generate 3D coordinates in a sphere
                theta = random.uniform(0, 2 * math.pi)
                phi = random.uniform(0, math.pi)
                radius = random.uniform(3, 8)

                x = radius * math.sin(phi) * math.cos(theta)
                y = radius * math.sin(phi) * math.sin(theta)
                z = radius * math.cos(phi)

                nodes_data.append(
                    {
                        "id": i + 1,
                        "title": node["title"],
                        "category": node["category"],
                        "content": node["content"],
                        "x": x,
                        "y": y,
                        "z": z,
                        "connections": random.randint(1, 4),
                        "importance": random.uniform(0.3, 1.0),
                    }
                )

        if not nodes_data:
            return cast(
                ToolResponse,
                {
                    "success": False,
                    "error": "No data found in database tables (technical_decisions, project_structure, memories) and failed to generate sample data",
                    "category": "NO_DATA_ERROR",
                },
            )

        # Generate semantic links if requested
        edges_data = []
        if include_semantic_links and len(nodes_data) > 1:
            # Use sophisticated semantic connection calculation
            edges_data = _calculate_semantic_connections(nodes_data)

        # Create timestamped filename
        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        filename = f"graph_3d_{timestamp}.html"
        file_path = os.path.join(output_dir, filename)

        # Generate 3D HTML visualization
        html_content = _generate_3d_html_visualization(
            nodes_data, edges_data, color_scheme, camera_position, animation_enabled
        )

        # Write to file
        with open(file_path, "w", encoding="utf-8") as f:
            f.write(html_content)

        # Convert to file:// URL for clickable link
        file_url = f"file:///{file_path.replace(os.sep, '/')}"

        return cast(
            ToolResponse,
            {
                "success": True,
                "file_path": file_path,
                "file_url": file_url,
                "stats": {
                    "nodes": len(nodes_data),
                    "edges": len(edges_data),
                    "dimensions": 3,
                    "categories": len(set(node["category"] for node in nodes_data)),
                },
                "3d_features": [
                    "WebGL Hardware Acceleration",
                    "Real-time Lighting & Shadows",
                    "Interactive Camera Controls",
                    "Animated Particle Effects",
                    f"Color Scheme: {color_scheme.title()}",
                    f"Camera: {camera_position.title()}",
                    f"Animation: {'Enabled' if animation_enabled else 'Disabled'}",
                ],
                "export_formats": export_formats,
                "instructions": [
                    "🖱️ Mouse: Orbit camera around the scene",
                    "🔍 Scroll: Zoom in/out",
                    "⌨️ Arrow Keys: Fine camera control",
                    "🎮 R: Reset camera position",
                    "💡 L: Toggle lighting effects",
                ],
            },
        )

    except Exception as e:
        return cast(
            ToolResponse,
            {
                "success": False,
                "error": f"Failed to create 3D visualization: {str(e)}",
                "category": "3D_GENERATION_ERROR",
                "details": {"table_name": table_name, "output_path": output_path},
            },
        )


def _generate_3d_html_visualization(
    nodes_data, edges_data, color_scheme, camera_position, animation_enabled
):
    """Generate Three.js-based 3D HTML visualization."""

    # Color schemes for 3D
    color_schemes = {
        "professional": {
            "background": "#0a0a0a",
            "node": "#4a90e2",
            "edge": "#cccccc",
            "text": "#ffffff",
            "accent": "#ff6b6b",
        },
        "vibrant": {
            "background": "#1a1a2e",
            "node": "#ff6b9d",
            "edge": "#00ffff",
            "text": "#ffffff",
            "accent": "#ffff00",
        },
        "neon": {
            "background": "#000000",
            "node": "#00ff41",
            "edge": "#ff00ff",
            "text": "#00ffff",
            "accent": "#ff4500",
        },
        "cosmic": {
            "background": "#0c0c1a",
            "node": "#8a2be2",
            "edge": "#ffd700",
            "text": "#ffffff",
            "accent": "#ff69b4",
        },
    }

    colors = color_schemes.get(color_scheme, color_schemes["professional"])

    # Convert data to JavaScript format
    nodes_js = json.dumps(nodes_data)
    edges_js = json.dumps(edges_data)

    html_template = f"""
<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>3D Knowledge Graph - SQLite Memory Bank</title>
    <script src="https://cdnjs.cloudflare.com/ajax/libs/three.js/r128/three.min.js"></script>
    <script src="https://cdn.jsdelivr.net/npm/three@0.128.0/examples/js/controls/OrbitControls.js"></script>
    <style>
        body {{
            margin: 0;
            padding: 0;
            background: {colors["background"]};
            color: {colors["text"]};
            font-family: 'Segoe UI', Tahoma, Geneva, Verdana, sans-serif;
            overflow: hidden;
        }}

        #container {{
            width: 100vw;
            height: 100vh;
            position: relative;
        }}

        #info {{
            position: absolute;
            top: 20px;
            left: 20px;
            z-index: 1000;
            background: rgba(0, 0, 0, 0.8);
            padding: 15px;
            border-radius: 10px;
            border: 1px solid {colors["accent"]};
            max-width: 300px;
        }}

        #controls {{
            position: absolute;
            bottom: 20px;
            left: 20px;
            z-index: 1000;
            background: rgba(0, 0, 0, 0.8);
            padding: 15px;
            border-radius: 10px;
            border: 1px solid {colors["accent"]};
        }}

        .control-button {{
            background: {colors["node"]};
            color: {colors["text"]};
            border: none;
            padding: 8px 16px;
            margin: 4px;
            border-radius: 5px;
            cursor: pointer;
            font-size: 12px;
        }}

        .control-button:hover {{
            background: {colors["accent"]};
            transform: translateY(-1px);
            box-shadow: 0 4px 8px rgba(0,0,0,0.3);
        }}

        /* Enhanced Modal Styles */
        .node-detail {{
            margin: 15px 0;
            padding: 10px;
            background: rgba(255, 255, 255, 0.05);
            border-radius: 8px;
            border-left: 3px solid {colors["accent"]};
        }}

        .node-detail-label {{
            font-weight: bold;
            color: {colors["accent"]};
            font-size: 12px;
            text-transform: uppercase;
            letter-spacing: 1px;
            margin-bottom: 5px;
        }}

        .node-detail-value {{
            color: #e0e0e0;
            line-height: 1.4;
        }}

        .category-badge {{
            display: inline-block;
            padding: 4px 12px;
            background: linear-gradient(135deg, {colors["node"]} 0%, {colors["edge"]} 100%);
            color: white;
            border-radius: 12px;
            font-size: 11px;
            font-weight: bold;
            text-transform: uppercase;
            letter-spacing: 0.5px;
        }}

        .importance-bar {{
            width: 100%;
            height: 6px;
            background: rgba(255, 255, 255, 0.1);
            border-radius: 3px;
            margin-top: 5px;
            overflow: hidden;
        }}

        .importance-fill {{
            height: 100%;
            background: linear-gradient(90deg, #ff6b6b 0%, #feca57 50%, #48dbfb 100%);
            border-radius: 3px;
            transition: width 0.3s ease;
        }}

        #stats {{
            position: absolute;
            top: 20px;
            right: 20px;
            z-index: 1000;
            background: rgba(0, 0, 0, 0.8);
            padding: 15px;
            border-radius: 10px;
            border: 1px solid {colors["accent"]};
        }}

        #filters {{
            position: absolute;
            bottom: 20px;
            right: 20px;
            z-index: 1000;
            background: rgba(0, 0, 0, 0.8);
            padding: 15px;
            border-radius: 10px;
            border: 1px solid {colors["accent"]};
            max-width: 250px;
        }}

        .filter-category {{
            margin: 5px 0;
        }}

        .filter-checkbox {{
            margin-right: 8px;
        }}

        .filter-label {{
            font-size: 12px;
            cursor: pointer;
        }}

        .node-info {{
            position: absolute;
            background: rgba(0, 0, 0, 0.9);
            padding: 10px;
            border-radius: 5px;
            border: 1px solid {colors["accent"]};
            pointer-events: none;
            font-size: 12px;
            display: none;
            z-index: 1001;
        }}

        /* Modal Styles */
        .modal {{
            display: none;
            position: fixed;
            z-index: 2000;
            left: 0;
            top: 0;
            width: 100%;
            height: 100%;
            background-color: rgba(0, 0, 0, 0.8);
            backdrop-filter: blur(5px);
        }}

        .modal-content {{
            background: linear-gradient(135deg, {colors["background"]}, #1a1a3a);
            margin: 5% auto;
            padding: 30px;
            border: 2px solid {colors["accent"]};
            border-radius: 15px;
            width: 80%;
            max-width: 600px;
            max-height: 80vh;
            overflow-y: auto;
            box-shadow: 0 10px 30px rgba(0, 0, 0, 0.5);
            position: relative;
        }}

        .modal-header {{
            display: flex;
            justify-content: space-between;
            align-items: center;
            margin-bottom: 20px;
            padding-bottom: 15px;
            border-bottom: 2px solid {colors["accent"]};
        }}

        .modal-title {{
            font-size: 24px;
            font-weight: bold;
            color: {colors["accent"]};
            margin: 0;
        }}

        .close-btn {{
            color: {colors["text"]};
            font-size: 28px;
            font-weight: bold;
            cursor: pointer;
            background: none;
            border: none;
            padding: 0;
            width: 30px;
            height: 30px;
            display: flex;
            align-items: center;
            justify-content: center;
            border-radius: 50%;
            transition: background-color 0.3s;
        }}

        .close-btn:hover {{
            background-color: {colors["accent"]};
            color: {colors["background"]};
        }}

        .modal-body {{
            line-height: 1.6;
        }}

        .node-detail {{
            margin: 15px 0;
            padding: 15px;
            background: rgba(255, 255, 255, 0.05);
            border-radius: 8px;
            border-left: 4px solid {colors["accent"]};
        }}

        .node-detail-label {{
            font-weight: bold;
            color: {colors["accent"]};
            font-size: 14px;
            margin-bottom: 5px;
        }}

        .node-detail-value {{
            color: {colors["text"]};
            font-size: 16px;
        }}

        .category-badge {{
            display: inline-block;
            padding: 4px 12px;
            background: {colors["node"]};
            color: {colors["text"]};
            border-radius: 20px;
            font-size: 12px;
            font-weight: bold;
            text-transform: uppercase;
        }}

        .importance-bar {{
            width: 100%;
            height: 8px;
            background: rgba(255, 255, 255, 0.2);
            border-radius: 4px;
            overflow: hidden;
            margin-top: 5px;
        }}

        .importance-fill {{
            height: 100%;
            background: linear-gradient(90deg, {colors["accent"]}, {colors["node"]});
            border-radius: 4px;
            transition: width 0.3s ease;
        }}

        .content-preview {{
            max-height: 150px;
            overflow-y: auto;
            line-height: 1.5;
            color: #cccccc;
        }}

        .node-actions {{
            margin-top: 20px;
            display: flex;
            gap: 10px;
            flex-wrap: wrap;
        }}

        .action-btn {{
            background: linear-gradient(135deg, {colors["accent"]}, {colors["node"]});
            color: {colors["text"]};
            border: none;
            padding: 8px 15px;
            border-radius: 20px;
            cursor: pointer;
            font-size: 12px;
            font-weight: bold;
            transition: all 0.3s ease;
            flex: 1;
            min-width: 120px;
        }}

        .action-btn:hover {{
            transform: translateY(-2px);
            box-shadow: 0 4px 12px rgba(255, 105, 180, 0.3);
        }}
    </style>
</head>
<body>
    <div id="container">
        <div id="info">
            <h3>🌐 3D Knowledge Graph</h3>
            <p><strong>Color Scheme:</strong> {color_scheme.title()}</p>
            <p><strong>Camera:</strong> {camera_position.title()}</p>
            <p><strong>Animation:</strong> {'Enabled' if animation_enabled else 'Disabled'}</p>
        </div>

        <div id="controls">
            <h4>🎮 Controls</h4>
            <button class="control-button" onclick="resetCamera()">🔄 Reset View</button>
            <button class="control-button" onclick="toggleAnimation()">⏯️ Animation</button>
            <button class="control-button" onclick="toggleLighting()">💡 Lighting</button>
            <button class="control-button" onclick="centerOnNodes()">🎯 Center</button>
            <button class="control-button" id="restoreBtn" onclick="restoreAllNodes()" style="display: none; background: #ff6b6b;">🔄 Show All Nodes</button>
        </div>

        <div id="stats">
            <h4>📊 Statistics</h4>
            <p><strong>Nodes:</strong> <span id="nodeCount">0</span></p>
            <p><strong>Edges:</strong> <span id="edgeCount">0</span></p>
            <p><strong>FPS:</strong> <span id="fps">0</span></p>
        </div>

        <div id="filters">
            <h4>🔍 Filters</h4>
            <div class="filter-category">
                <input type="checkbox" id="filter-all" class="filter-checkbox" checked>
                <label for="filter-all" class="filter-label">Show All</label>
            </div>
            <div id="category-filters"></div>
            <div class="filter-category">
                <label class="filter-label">Min Importance:</label>
                <input type="range" id="importance-slider" min="0" max="1" step="0.1" value="0" style="width: 100%; margin-top: 5px;">
                <span id="importance-value">0.0</span>
            </div>
        </div>

        <div class="node-info" id="nodeTooltip"></div>

        <!-- Modal for detailed node information -->
        <div id="nodeModal" class="modal">
            <div class="modal-content">
                <div class="modal-header">
                    <h2 class="modal-title" id="modalTitle">Node Details</h2>
                    <button class="close-btn" id="closeModal" onclick="closeModal()">&times;</button>
                </div>
                <div class="modal-body" id="modalBody">
                    <!-- Dynamic content will be inserted here -->
                </div>
            </div>
        </div>
                </div>
            </div>
        </div>
    </div>

    <script>
        // Global variables
        let scene, camera, renderer, controls;
        let nodes = {nodes_js};
        let edges = {edges_js};
        let nodeObjects = [];
        let edgeObjects = [];
        let animationEnabled = {str(animation_enabled).lower()};
        let lightingEnabled = true;
        let ambientLight, pointLight;
        let raycaster, mouse;
        let activeFilters = new Set();
        let minImportance = 0.0;
        let selectedNode = null;

        // Initialize the 3D scene
        function init() {{
            // Create scene
            scene = new THREE.Scene();
            scene.background = new THREE.Color('{colors["background"]}');

            // Create camera
            const aspect = window.innerWidth / window.innerHeight;
            if ('{camera_position}' === 'orthographic') {{
                const frustumSize = 20;
                camera = new THREE.OrthographicCamera(
                    frustumSize * aspect / -2, frustumSize * aspect / 2,
                    frustumSize / 2, frustumSize / -2,
                    0.1, 1000
                );
            }} else {{
                camera = new THREE.PerspectiveCamera(75, aspect, 0.1, 1000);
            }}
            camera.position.set(10, 10, 10);

            // Create renderer
            renderer = new THREE.WebGLRenderer({{ antialias: true }});
            renderer.setSize(window.innerWidth, window.innerHeight);
            renderer.shadowMap.enabled = true;
            renderer.shadowMap.type = THREE.PCFSoftShadowMap;
            document.getElementById('container').appendChild(renderer.domElement);

            // Add lighting
            setupLighting();

            // Create controls
            controls = new THREE.OrbitControls(camera, renderer.domElement);
            controls.enableDamping = true;
            controls.dampingFactor = 0.05;
            controls.screenSpacePanning = false;
            controls.minDistance = 0.1;  // Allow much closer zoom
            controls.maxDistance = 1000; // Allow much farther zoom
            controls.enableZoom = true;
            controls.zoomSpeed = 1.0;

            // Create raycaster for mouse interaction
            raycaster = new THREE.Raycaster();
            mouse = new THREE.Vector2();

            // Create 3D objects
            createNodes();
            createEdges();

            // Add event listeners
            window.addEventListener('resize', onWindowResize);
            renderer.domElement.addEventListener('mousemove', onMouseMove, false);
            renderer.domElement.addEventListener('click', onMouseClick, false);

            // Modal close listeners (with null checks)
            const closeModalBtn = document.getElementById('closeModal');
            if (closeModalBtn) {{
                closeModalBtn.addEventListener('click', closeModal);
            }}

            const nodeModal = document.getElementById('nodeModal');
            if (nodeModal) {{
                nodeModal.addEventListener('click', (e) => {{
                    if (e.target.id === 'nodeModal') {{
                        closeModal();
                    }}
                }});
            }}

            // ESC key to close modal
            document.addEventListener('keydown', (e) => {{
                if (e.key === 'Escape') {{
                    closeModal();
                }}
            }});

            // Initialize filters (with null checks)
            try {{
                initializeFilters();
            }} catch (e) {{
                console.log('Filters not available:', e);
            }}

            // Update initial stats (with null checks)
            const nodeCountEl = document.getElementById('nodeCount');
            const edgeCountEl = document.getElementById('edgeCount');

            if (nodeCountEl) {{
                nodeCountEl.textContent = nodes.length.toString();
            }}
            if (edgeCountEl) {{
                edgeCountEl.textContent = edges.length.toString();
            }}

            console.log('🌌 3D Knowledge Graph initialized with', nodes.length, 'nodes and', edges.length, 'edges');

            // Start animation loop
            animate();
        }}

        function setupLighting() {{
            // Ambient light
            ambientLight = new THREE.AmbientLight(0x404040, 1.2);  // Much brighter ambient light
            scene.add(ambientLight);

            // Point light
            pointLight = new THREE.PointLight('{colors["accent"]}', 1.0, 100);
            pointLight.position.set(10, 10, 10);
            pointLight.castShadow = true;
            pointLight.shadow.mapSize.width = 2048;
            pointLight.shadow.mapSize.height = 2048;
            scene.add(pointLight);

            // Additional colored lights for atmosphere
            const coloredLight1 = new THREE.PointLight('{colors["node"]}', 0.5, 50);
            coloredLight1.position.set(-10, 5, -10);
            scene.add(coloredLight1);

            const coloredLight2 = new THREE.PointLight('{colors["edge"]}', 0.3, 30);
            coloredLight2.position.set(5, -5, 15);
            scene.add(coloredLight2);
        }}

        function createNodes() {{
            nodes.forEach((node, index) => {{
                // Create node geometry based on importance
                const radius = 0.3 + (node.importance * 0.5);
                const geometry = new THREE.SphereGeometry(radius, 16, 16);

                // Create material with semantic category-based colors - MUCH BRIGHTER
                const color = node.color || '{colors["node"]}';  // Use color from semantic clustering
                const material = new THREE.MeshPhongMaterial({{
                    color: color,
                    transparent: false,  // Disable transparency for better visibility
                    opacity: 1.0,  // Full opacity
                    shininess: 30,   // Reduced shininess for better visibility
                    emissive: 0x333333  // Much stronger emissive glow for visibility
                }});

                const sphere = new THREE.Mesh(geometry, material);
                sphere.position.set(node.x, node.y, node.z);
                sphere.castShadow = true;
                sphere.receiveShadow = true;

                // Store node data for interaction
                sphere.userData = {{ nodeData: node, nodeIndex: index }};

                scene.add(sphere);
                nodeObjects.push(sphere);

                // Add text label
                createTextLabel(node.title, sphere.position, 0.8);
            }});
        }}

        function createEdges() {{
            edges.forEach(edge => {{
                const sourceNode = nodes.find(n => n.id === edge.source);
                const targetNode = nodes.find(n => n.id === edge.target);

                if (sourceNode && targetNode) {{
                    const points = [
                        new THREE.Vector3(sourceNode.x, sourceNode.y, sourceNode.z),
                        new THREE.Vector3(targetNode.x, targetNode.y, targetNode.z)
                    ];

                    const geometry = new THREE.BufferGeometry().setFromPoints(points);
                    const material = new THREE.LineBasicMaterial({{
                        color: '{colors["edge"]}',
                        transparent: true,
                        opacity: edge.strength || 0.6
                    }});

                    const line = new THREE.Line(geometry, material);

                    // CRITICAL: Store edge data for filtering and relationship display
                    line.userData = {{
                        edgeData: edge
                    }};

                    scene.add(line);
                    edgeObjects.push(line);
                }}
            }});
        }}

        function createTextLabel(text, position, size) {{
            const canvas = document.createElement('canvas');
            const context = canvas.getContext('2d');
            canvas.width = 256;
            canvas.height = 64;

            context.fillStyle = '{colors["text"]}';
            context.font = '16px Arial';
            context.textAlign = 'center';
            context.fillText(text, 128, 32);

            const texture = new THREE.CanvasTexture(canvas);
            const material = new THREE.SpriteMaterial({{ map: texture }});
            const sprite = new THREE.Sprite(material);

            sprite.position.copy(position);
            sprite.position.y += size;
            sprite.scale.set(2, 0.5, 1);

            scene.add(sprite);
        }}

        function animate() {{
            requestAnimationFrame(animate);

            // Animation effects
            if (animationEnabled) {{
                const time = Date.now() * 0.001;

                // Rotate nodes gently
                nodeObjects.forEach((node, index) => {{
                    node.rotation.y = time * 0.2 + index * 0.1;

                    // Gentle floating animation
                    const originalY = nodes[index].y;
                    node.position.y = originalY + Math.sin(time + index) * 0.1;
                }});

                // Move lights for dynamic lighting
                if (pointLight) {{
                    pointLight.position.x = Math.sin(time * 0.5) * 15;
                    pointLight.position.z = Math.cos(time * 0.5) * 15;
                }}
            }}

            controls.update();
            renderer.render(scene, camera);

            // Update FPS counter
            updateFPS();
        }}

        let lastTime = 0;
        let frameCount = 0;
        function updateFPS() {{
            frameCount++;
            const currentTime = Date.now();
            if (currentTime - lastTime >= 1000) {{
                document.getElementById('fps').textContent = frameCount;
                frameCount = 0;
                lastTime = currentTime;
            }}
        }}

        function onMouseMove(event) {{
            mouse.x = (event.clientX / window.innerWidth) * 2 - 1;
            mouse.y = -(event.clientY / window.innerHeight) * 2 + 1;

            raycaster.setFromCamera(mouse, camera);
            const intersects = raycaster.intersectObjects(nodeObjects);

            const tooltip = document.getElementById('nodeTooltip');

            if (intersects.length > 0) {{
                const node = intersects[0].object.userData.nodeData;
                tooltip.innerHTML = `
                    <strong>${{node.title}}</strong><br>
                    Category: ${{node.category}}<br>
                    Importance: ${{node.importance.toFixed(2)}}<br>
                    Connections: ${{node.connections}}<br>
                    <em>${{node.content.substring(0, 50)}}...</em><br>
                    <small>💡 Click for details</small>
                `;
                tooltip.style.left = event.clientX + 10 + 'px';
                tooltip.style.top = event.clientY + 10 + 'px';
                tooltip.style.display = 'block';

                // Highlight hovered node
                intersects[0].object.material.emissive.setHex(0x333333);
            }} else {{
                tooltip.style.display = 'none';
                // Remove highlight from all nodes
                nodeObjects.forEach(nodeObj => {{
                    nodeObj.material.emissive.setHex(0x000000);
                }});
            }}
        }}

        function onMouseClick(event) {{
            mouse.x = (event.clientX / window.innerWidth) * 2 - 1;
            mouse.y = -(event.clientY / window.innerHeight) * 2 + 1;

            raycaster.setFromCamera(mouse, camera);
            const intersects = raycaster.intersectObjects(nodeObjects);

            if (intersects.length > 0) {{
                const node = intersects[0].object.userData.nodeData;
                showNodeModal(node);

                // Highlight selected node
                nodeObjects.forEach(nodeObj => {{
                    nodeObj.material.emissive.setHex(0x000000);
                }});
                intersects[0].object.material.emissive.setHex(0x444444);
                selectedNode = intersects[0].object;
            }}
        }}

        function showNodeModal(node) {{
            const modal = document.getElementById('nodeModal');
            const title = document.getElementById('modalTitle');
            const body = document.getElementById('modalBody');

            title.textContent = node.title;

            // Create detailed modal content
            body.innerHTML = `
                <div class="node-detail">
                    <div class="node-detail-label">Category</div>
                    <div class="node-detail-value">
                        <span class="category-badge">${{node.category}}</span>
                    </div>
                </div>

                <div class="node-detail">
                    <div class="node-detail-label">Importance Score</div>
                    <div class="node-detail-value">
                        ${{node.importance.toFixed(2)}} / 1.00
                        <div class="importance-bar">
                            <div class="importance-fill" style="width: ${{node.importance * 100}}%"></div>
                        </div>
                    </div>
                </div>

                <div class="node-detail">
                    <div class="node-detail-label">Connections</div>
                    <div class="node-detail-value">${{node.connections}} related nodes</div>
                </div>

                <div class="node-detail">
                    <div class="node-detail-label">3D Position</div>
                    <div class="node-detail-value">
                        X: ${{node.x.toFixed(1)}}, Y: ${{node.y.toFixed(1)}}, Z: ${{node.z.toFixed(1)}}
                    </div>
                </div>

                <div class="node-detail">
                    <div class="node-detail-label">Description</div>
                    <div class="node-detail-value">${{node.content}}</div>
                </div>

                <div class="node-detail">
                    <div class="node-detail-label">Actions</div>
                    <div class="node-detail-value">
                        <button class="control-button" onclick="focusOnNode(${{node.id}})">🎯 Focus Camera</button>
                        <button class="control-button" onclick="highlightConnections(${{node.id}})">🔗 Show Connections</button>
                        <button class="control-button" onclick="filterByCategory('${{node.category}}')">🔍 Filter Category</button>
                    </div>
                </div>
            `;

            modal.style.display = 'block';
        }}

        function closeModal() {{
            const modal = document.getElementById('nodeModal');
            modal.style.display = 'none';

            // Remove selection highlight
            if (selectedNode) {{
                selectedNode.material.emissive.setHex(0x000000);
                selectedNode = null;
            }}
        }}

        function initializeFilters() {{
            // Get unique categories
            const categories = [...new Set(nodes.map(node => node.category))];
            const filterContainer = document.getElementById('category-filters');

            // Create category checkboxes with color indicators
            categories.forEach(category => {{
                const div = document.createElement('div');
                div.className = 'filter-category';

                // Get color for this category
                const categoryColors = {{
                    'architecture': '#e74c3c',
                    'deployment': '#3498db',
                    'database': '#2ecc71',
                    'api_design': '#f39c12',
                    'testing': '#9b59b6',
                    'development': '#1abc9c',
                    'configuration': '#34495e',
                    'documentation': '#f1c40f',
                    'performance': '#e67e22',
                    'security': '#c0392b',
                    'error_handling': '#8e44ad',
                    'workflow': '#16a085',
                    'general': '#95a5a6'
                }};
                const color = categoryColors[category] || '#95a5a6';

                div.innerHTML = `
                    <input type="checkbox" id="filter-${{category}}" class="filter-checkbox" checked>
                    <span class="color-indicator" style="background-color: ${{color}}; width: 12px; height: 12px; display: inline-block; border-radius: 50%; margin-right: 6px; vertical-align: middle;"></span>
                    <label for="filter-${{category}}" class="filter-label">${{category}} (${{nodes.filter(n => n.category === category).length}})</label>
                `;
                filterContainer.appendChild(div);

                // Add event listener
                const checkbox = div.querySelector(`#filter-${{category}}`);
                checkbox.addEventListener('change', () => updateFilters());
            }});

            // Add event listeners for other filters
            document.getElementById('filter-all').addEventListener('change', toggleAllFilters);
            document.getElementById('importance-slider').addEventListener('input', updateImportanceFilter);

            // Initialize active filters
            activeFilters = new Set(categories);
        }}

        function toggleAllFilters() {{
            const showAll = document.getElementById('filter-all').checked;
            const categoryCheckboxes = document.querySelectorAll('.filter-checkbox:not(#filter-all)');

            categoryCheckboxes.forEach(checkbox => {{
                checkbox.checked = showAll;
            }});

            updateFilters();
        }}

        function updateImportanceFilter() {{
            const slider = document.getElementById('importance-slider');
            const valueSpan = document.getElementById('importance-value');
            minImportance = parseFloat(slider.value);
            valueSpan.textContent = minImportance.toFixed(1);
            updateFilters();
        }}

        function updateFilters() {{
            // Don't update filters if we're in connection mode (unless called from restoreAllNodes)
            if (window.connectionModeActive && !window.restoreInProgress) {{
                return;
            }}

            // Update active filters set
            activeFilters.clear();
            const categoryCheckboxes = document.querySelectorAll('.filter-checkbox:not(#filter-all)');
            categoryCheckboxes.forEach(checkbox => {{
                if (checkbox.checked) {{
                    const category = checkbox.id.replace('filter-', '');
                    activeFilters.add(category);
                }}
            }});

            // Apply filters to nodes - HIDE filtered nodes completely
            nodeObjects.forEach((nodeObj, index) => {{
                const node = nodes[index];
                const isVisible = activeFilters.has(node.category) && node.importance >= minImportance;

                if (isVisible) {{
                    // Show node - full visibility
                    nodeObj.visible = true;
                    nodeObj.material.opacity = 1.0;
                }} else {{
                    // Hide filtered node completely
                    nodeObj.visible = false;
                }}
            }});

            // Apply filters to edges - HIDE edges to hidden nodes
            edgeObjects.forEach(edgeObj => {{
                const edge = edgeObj.userData?.edgeData;
                if (edge) {{
                    const sourceNode = nodes.find(n => n.id === edge.source);
                    const targetNode = nodes.find(n => n.id === edge.target);

                    const sourceVisible = activeFilters.has(sourceNode.category) && sourceNode.importance >= minImportance;
                    const targetVisible = activeFilters.has(targetNode.category) && targetNode.importance >= minImportance;

                    if (sourceVisible && targetVisible) {{
                        edgeObj.visible = true;
                        edgeObj.material.opacity = edge.strength || 0.6;
                    }} else {{
                        edgeObj.visible = false;
                    }}
                }}
            }});

            // Update stats
            const visibleNodes = nodes.filter(node =>
                activeFilters.has(node.category) && node.importance >= minImportance
            ).length;
            document.getElementById('nodeCount').textContent = `${{visibleNodes}}/${{nodes.length}}`;
        }}

        function filterByCategory(category) {{
            // Uncheck all except the selected category
            const categoryCheckboxes = document.querySelectorAll('.filter-checkbox:not(#filter-all)');
            categoryCheckboxes.forEach(checkbox => {{
                const checkboxCategory = checkbox.id.replace('filter-', '');
                checkbox.checked = (checkboxCategory === category);
            }});

            // Uncheck "Show All"
            document.getElementById('filter-all').checked = false;

            updateFilters();
            closeModal();
        }}

        function focusOnNode(nodeId) {{
            const node = nodes.find(n => n.id === nodeId);
            if (node) {{
                const targetPosition = new THREE.Vector3(node.x, node.y, node.z);

                // Animate camera to focus on the node
                const distance = 5;
                const newCameraPosition = targetPosition.clone();
                newCameraPosition.z += distance;

                // Smooth camera transition
                animateCameraTo(newCameraPosition, targetPosition);
            }}
            closeModal();
        }}

        function restoreAllNodes() {{
            // Reset connection mode
            window.connectionModeActive = false;
            window.restoreInProgress = true; // Allow filters to work during restore

            // Hide the restore button
            const restoreBtn = document.getElementById('restoreBtn');
            if (restoreBtn) {{
                restoreBtn.style.display = 'none';
            }}

            // Remove all relationship labels
            edgeObjects.forEach(edgeObj => {{
                if (edgeObj.userData.labelSprite) {{
                    scene.remove(edgeObj.userData.labelSprite);
                    edgeObj.userData.labelSprite = null;
                }}
            }});

            // Restore all nodes to full visibility
            nodeObjects.forEach(nodeObj => {{
                nodeObj.visible = true;
                nodeObj.material.opacity = 1.0;
                nodeObj.material.emissive.setHex(0x333333); // Restore normal emissive glow
            }});

            // Restore all edges to normal state
            edgeObjects.forEach(edgeObj => {{
                edgeObj.visible = true;
                edgeObj.material.opacity = 0.6; // Normal edge opacity
                edgeObj.material.color.setHex(0x888888); // Normal gray color
            }});

            // Apply current filters to determine visibility
            updateFilters();
            window.restoreInProgress = false; // Reset flag
        }}

        function animateCameraTo(position, target) {{
            const startPosition = camera.position.clone();
            const startTarget = controls.target.clone();
            const duration = 1000; // 1 second
            const startTime = Date.now();

            function animate() {{
                const elapsed = Date.now() - startTime;
                const progress = Math.min(elapsed / duration, 1);
                const easeProgress = 1 - Math.pow(1 - progress, 3); // Ease out cubic

                camera.position.lerpVectors(startPosition, position, easeProgress);
                controls.target.lerpVectors(startTarget, target, easeProgress);
                controls.update();

                if (progress < 1) {{
                    requestAnimationFrame(animate);
                }}
            }}

            animate();
        }}

        function onWindowResize() {{
            const aspect = window.innerWidth / window.innerHeight;

            if (camera.isPerspectiveCamera) {{
                camera.aspect = aspect;
            }} else {{
                const frustumSize = 20;
                camera.left = frustumSize * aspect / -2;
                camera.right = frustumSize * aspect / 2;
            }}

            camera.updateProjectionMatrix();
            renderer.setSize(window.innerWidth, window.innerHeight);
        }}

        // Control functions
        function resetCamera() {{
            camera.position.set(10, 10, 10);
            controls.reset();
        }}

        function toggleAnimation() {{
            animationEnabled = !animationEnabled;
        }}

        function toggleLighting() {{
            lightingEnabled = !lightingEnabled;
            // Fix: When lighting is disabled, increase ambient light instead of making it invisible
            if (lightingEnabled) {{
                ambientLight.intensity = 1.2;  // Bright ambient
                pointLight.intensity = 1.0;
                pointLight.visible = true;
            }} else {{
                ambientLight.intensity = 2.0;  // Even brighter ambient when point light is off
                pointLight.intensity = 0.0;
                pointLight.visible = false;
            }}
        }}

        function centerOnNodes() {{
            const box = new THREE.Box3();
            nodeObjects.forEach(node => box.expandByObject(node));
            const center = box.getCenter(new THREE.Vector3());
            const size = box.getSize(new THREE.Vector3());

            const maxDim = Math.max(size.x, size.y, size.z);
            const fov = camera.fov * (Math.PI / 180);
            const distance = maxDim / (2 * Math.tan(fov / 2)) * 1.5;

            camera.position.copy(center);
            camera.position.z += distance;
            camera.lookAt(center);
            controls.target.copy(center);
        }}

        // Keyboard controls
        document.addEventListener('keydown', (event) => {{
            switch(event.code) {{
                case 'KeyR':
                    resetCamera();
                    break;
                case 'KeyL':
                    toggleLighting();
                    break;
                case 'Space':
                    toggleAnimation();
                    event.preventDefault();
                    break;
                case 'Escape':
                    if (typeof closeModal === 'function') closeModal();
                    break;
            }}
        }});

        // Modal and utility functions

        function closeModal() {{
            const modal = document.getElementById('nodeModal');
            if (modal) {{
                modal.style.display = 'none';
            }}
        }}

        function showNodeModal(node) {{
            const modal = document.getElementById('nodeModal');
            const title = document.getElementById('modalTitle');
            const body = document.getElementById('modalBody');

            if (modal && title && body) {{
                title.textContent = node.title;

                // Create detailed modal content with meaningful information
                body.innerHTML = `
                    <div class="node-detail">
                        <div class="node-detail-label">Category</div>
                        <div class="node-detail-value">
                            <span class="category-badge" style="background-color: ${{node.color || '#95a5a6'}}">${{node.category.replace('_', ' ').toUpperCase()}}</span>
                        </div>
                    </div>

                    <div class="node-detail">
                        <div class="node-detail-label">Relevance Score</div>
                        <div class="node-detail-value">
                            ${{node.importance.toFixed(2)}} / 1.00
                            <div class="importance-bar">
                                <div class="importance-fill" style="width: ${{node.importance * 100}}%"></div>
                            </div>
                        </div>
                    </div>

                    <div class="node-detail">
                        <div class="node-detail-label">Knowledge Network</div>
                        <div class="node-detail-value">${{node.connections}} related concepts</div>
                    </div>

                    <div class="node-detail">
                        <div class="node-detail-label">Source</div>
                        <div class="node-detail-value">${{node.source_table?.replace('_', ' ') || 'Unknown'}}</div>
                    </div>

                    <div class="node-detail">
                        <div class="node-detail-label">Content</div>
                        <div class="node-detail-value content-preview">${{node.content}}</div>
                    </div>

                    <div class="node-actions">
                        <button onclick="filterByCategory('${{node.category}}')" class="action-btn">
                            🔍 Show Similar
                        </button>
                        <button onclick="highlightConnections('${{node.id}}')" class="action-btn">
                            🔗 Show Connections
                        </button>
                        <button onclick="focusOnNode('${{node.id}}')" class="action-btn">
                            🎯 Focus View
                        </button>
                    </div>
                `;

                modal.style.display = 'block';
            }}
        }}

        function onMouseMove(event) {{
            if (!mouse) mouse = new THREE.Vector2();
            if (!raycaster) raycaster = new THREE.Raycaster();

            mouse.x = (event.clientX / window.innerWidth) * 2 - 1;
            mouse.y = -(event.clientY / window.innerHeight) * 2 + 1;

            raycaster.setFromCamera(mouse, camera);
            const intersects = raycaster.intersectObjects(nodeObjects);

            const tooltip = document.getElementById('nodeTooltip');

            if (intersects.length > 0 && tooltip) {{
                const node = intersects[0].object.userData.nodeData;
                tooltip.innerHTML = `
                    <strong>${{node.title}}</strong><br>
                    Category: ${{node.category}}<br>
                    Importance: ${{node.importance.toFixed(2)}}<br>
                    Connections: ${{node.connections}}<br>
                    <em>${{node.content.substring(0, 50)}}...</em><br>
                    <small>💡 Click for details</small>
                `;
                tooltip.style.left = event.clientX + 10 + 'px';
                tooltip.style.top = event.clientY + 10 + 'px';
                tooltip.style.display = 'block';

                // Highlight hovered node
                intersects[0].object.material.emissive.setHex(0x333333);
            }} else {{
                if (tooltip) tooltip.style.display = 'none';
                // Remove highlight from all nodes
                nodeObjects.forEach(nodeObj => {{
                    nodeObj.material.emissive.setHex(0x000000);
                }});
            }}
        }}

        function onMouseClick(event) {{
            if (!mouse) mouse = new THREE.Vector2();
            if (!raycaster) raycaster = new THREE.Raycaster();

            mouse.x = (event.clientX / window.innerWidth) * 2 - 1;
            mouse.y = -(event.clientY / window.innerHeight) * 2 + 1;

            raycaster.setFromCamera(mouse, camera);
            const intersects = raycaster.intersectObjects(nodeObjects);

            if (intersects.length > 0) {{
                const node = intersects[0].object.userData.nodeData;
                showNodeModal(node);

                // Highlight selected node
                nodeObjects.forEach(nodeObj => {{
                    nodeObj.material.emissive.setHex(0x000000);
                }});
                intersects[0].object.material.emissive.setHex(0x444444);
            }}
        }}

        // Helper function to create text texture for relationship labels
        function createTextTexture(text, options = {{}}) {{
            const canvas = document.createElement('canvas');
            const context = canvas.getContext('2d');

            const fontSize = options.fontsize || 24;
            const fontFamily = options.fontFamily || 'Arial, sans-serif';
            const textColor = options.textColor || '#ffffff';
            const backgroundColor = options.backgroundColor || 'rgba(0,0,0,0.8)';

            context.font = `${{fontSize}}px ${{fontFamily}}`;
            const textMetrics = context.measureText(text);
            const textWidth = textMetrics.width;
            const textHeight = fontSize;

            canvas.width = Math.max(textWidth + 20, 100);
            canvas.height = textHeight + 16;

            // Set font again after resizing canvas
            context.font = `${{fontSize}}px ${{fontFamily}}`;
            context.textAlign = 'center';
            context.textBaseline = 'middle';

            // Draw background
            context.fillStyle = backgroundColor;
            context.fillRect(0, 0, canvas.width, canvas.height);

            // Draw text
            context.fillStyle = textColor;
            context.fillText(text, canvas.width / 2, canvas.height / 2);

            const texture = new THREE.CanvasTexture(canvas);
            texture.needsUpdate = true;
            return texture;
        }}

        function highlightConnections(nodeId) {{
            // Store current state for restoration
            window.connectionModeActive = true;
            window.lastSelectedNode = nodeId;

            // Show the restore button
            const restoreBtn = document.getElementById('restoreBtn');
            if (restoreBtn) {{
                restoreBtn.style.display = 'inline-block';
            }}

            // Hide all nodes first
            nodeObjects.forEach(nodeObj => {{
                nodeObj.visible = false;
                nodeObj.material.emissive.setHex(0x000000);
            }});

            // Hide all edges first
            edgeObjects.forEach(edgeObj => {{
                edgeObj.visible = false;
            }});

            // Show and highlight the selected node
            const targetNodeObj = nodeObjects.find(obj => obj.userData.nodeData.id === nodeId);
            if (targetNodeObj) {{
                targetNodeObj.visible = true;
                targetNodeObj.material.opacity = 1.0;
                targetNodeObj.material.emissive.setHex(0xff6b6b); // Bright red highlight
            }}

            // Find all connected nodes (add the selected node to the set too)
            const connectedNodeIds = new Set();
            connectedNodeIds.add(nodeId); // Include the selected node itself
            edges.forEach(edge => {{
                if (edge.source === nodeId) {{
                    connectedNodeIds.add(edge.target);
                }} else if (edge.target === nodeId) {{
                    connectedNodeIds.add(edge.source);
                }}
            }});

            // Show and highlight connected nodes
            nodeObjects.forEach(nodeObj => {{
                if (connectedNodeIds.has(nodeObj.userData.nodeData.id)) {{
                    nodeObj.visible = true;
                    nodeObj.material.opacity = 1.0;
                    nodeObj.material.emissive.setHex(0x4a90e2); // Blue highlight for connected nodes
                }}
            }});

            // Show ALL edges between visible nodes (not just edges to selected node)
            edgeObjects.forEach(edgeObj => {{
                const edge = edgeObj.userData?.edgeData;
                if (edge && connectedNodeIds.has(edge.source) && connectedNodeIds.has(edge.target)) {{
                    edgeObj.visible = true;
                    edgeObj.material.opacity = 1.0;

                    // Color coding: yellow for direct connections to selected node, white for inter-connections
                    if (edge.source === nodeId || edge.target === nodeId) {{
                        edgeObj.material.color.setHex(0xffff00); // Yellow for direct connections
                    }} else {{
                        edgeObj.material.color.setHex(0xffffff); // White for connections between connected nodes
                    }}

                    // Add relationship label if it doesn't exist
                    if (!edgeObj.userData.labelSprite && edge.relationship) {{
                        const labelTexture = createTextTexture(edge.relationship, {{
                            fontsize: 24,
                            backgroundColor: 'rgba(0,0,0,0.8)',
                            textColor: '#ffffff'
                        }});
                        const labelMaterial = new THREE.SpriteMaterial({{ map: labelTexture }});
                        const labelSprite = new THREE.Sprite(labelMaterial);

                        // Position label at edge midpoint
                        const sourcePos = nodeObjects.find(obj => obj.userData.nodeData.id === edge.source)?.position;
                        const targetPos = nodeObjects.find(obj => obj.userData.nodeData.id === edge.target)?.position;
                        if (sourcePos && targetPos) {{
                            labelSprite.position.set(
                                (sourcePos.x + targetPos.x) / 2,
                                (sourcePos.y + targetPos.y) / 2,
                                (sourcePos.z + targetPos.z) / 2
                            );
                            labelSprite.scale.set(5, 2, 1);
                            scene.add(labelSprite);
                            edgeObj.userData.labelSprite = labelSprite;
                        }}
                    }}
                }}
            }});

            closeModal();
        }}

        // Initialize the scene
        init();
    </script>
</body>
</html>
"""

    return html_template


def _collect_dashboard_data(
    database, dashboard_type: str, include_metrics: bool
) -> Dict[str, Any]:
    """Collect comprehensive data for dashboard visualization."""
    # Implementation for dashboard data collection
    return {
        "overview": {"total_tables": 0, "total_rows": 0},
        "timeline": [],
        "metrics": {} if include_metrics else None,
    }


def _generate_dashboard_html(
    data: Dict,
    dashboard_type: str,
    include_metrics: bool,
    real_time: bool,
    widgets: List[str],
) -> str:
    """Generate comprehensive dashboard HTML."""
    return "<html><body><h1>Dashboard Coming Soon</h1></body></html>"


def _generate_dashboard_assets(output_dir: Path, dashboard_type: str) -> None:
    """Generate supporting CSS and JS assets for dashboard."""


def _generate_exports(
    file_path: Path, graph_data: Dict, formats: List[str]
) -> List[str]:
    """Generate additional export formats."""
    export_paths = []
    # Implementation for generating exports
    return export_paths


def _export_to_json(graph_data: Dict, include_metadata: bool) -> str:
    """Export graph data to JSON format."""
    return json.dumps(graph_data, indent=2)


def _export_to_graphml(graph_data: Dict, include_metadata: bool) -> str:
    """Export graph data to GraphML format."""
    # Implementation for GraphML export
    return '<?xml version="1.0" encoding="UTF-8"?><graphml></graphml>'


def _export_to_gexf(graph_data: Dict, include_metadata: bool) -> str:
    """Export graph data to GEXF format."""
    # Implementation for GEXF export
    return '<?xml version="1.0" encoding="UTF-8"?><gexf></gexf>'


def _export_to_cytoscape(graph_data: Dict, include_metadata: bool) -> str:
    """Export graph data to Cytoscape format."""
    # Implementation for Cytoscape export
    return json.dumps({"nodes": [], "edges": []}, indent=2)


def _create_semantic_clusters(raw_data: List[Dict]) -> List[Dict]:
    """
    Create semantic clusters from raw database content using content analysis
    and intelligent categorization rather than random positioning.
    """
    import random
    import math

    if not raw_data:
        return []

    # Extract semantic topics from content using keyword analysis
    def extract_semantic_category(title: str, content: str) -> str:
        """Extract meaningful semantic category from content"""
        text = f"{title} {content}".lower()

        # Define semantic categories with keywords
        categories = {
            "deployment": [
                "deploy",
                "release",
                "pipeline",
                "ci/cd",
                "build",
                "publish",
                "version",
                "completion",
            ],
            "development": [
                "integration",
                "implementation",
                "coding",
                "refactor",
                "enhancement",
                "tool",
                "development",
            ],
            "testing": [
                "test",
                "testing",
                "validation",
                "verification",
                "quality",
                "coverage",
            ],
            "configuration": [
                "config",
                "setup",
                "environment",
                "settings",
                "installation",
                "server",
            ],
            "architecture": [
                "architecture",
                "design",
                "structure",
                "component",
                "module",
                "pattern",
                "framework",
                "system",
            ],
            "database": [
                "database",
                "sqlite",
                "memory",
                "storage",
                "schema",
                "query",
                "table",
            ],
            "api_design": [
                "api",
                "endpoint",
                "rest",
                "interface",
                "method",
                "function",
            ],
            "documentation": [
                "documentation",
                "docs",
                "guide",
                "instruction",
                "readme",
            ],
            "performance": [
                "performance",
                "optimization",
                "speed",
                "efficiency",
                "scalability",
            ],
            "security": [
                "security",
                "authentication",
                "authorization",
                "encryption",
                "vulnerability",
            ],
            "error_handling": [
                "error",
                "exception",
                "handling",
                "debugging",
                "troubleshoot",
            ],
            "workflow": ["workflow", "process", "procedure", "protocol", "methodology"],
        }

        # Score each category based on keyword matches
        category_scores = {}
        for category, keywords in categories.items():
            score = sum(1 for keyword in keywords if keyword in text)
            if score > 0:
                category_scores[category] = score

        # Return highest scoring category or fallback
        if category_scores:
            return max(category_scores.items(), key=lambda x: x[1])[0]
        else:
            return "general"

    # Categorize all nodes semantically
    categorized_nodes = {}
    for item in raw_data:
        category = extract_semantic_category(item["title"], item["content"])
        if category not in categorized_nodes:
            categorized_nodes[category] = []
        categorized_nodes[category].append(item)

    # Generate spatial clusters for each semantic category
    nodes_data = []

    # Define cluster centers in 3D space for visual separation
    category_colors = {
        "architecture": "#e74c3c",  # Red - core systems
        "deployment": "#3498db",  # Blue - operations
        "database": "#2ecc71",  # Green - data
        "api_design": "#f39c12",  # Orange - interfaces
        "testing": "#9b59b6",  # Purple - quality
        "development": "#1abc9c",  # Teal - coding
        "configuration": "#34495e",  # Dark gray - setup
        "documentation": "#f1c40f",  # Yellow - docs
        "performance": "#e67e22",  # Dark orange - optimization
        "security": "#c0392b",  # Dark red - security
        "error_handling": "#8e44ad",  # Dark purple - errors
        "workflow": "#16a085",  # Dark teal - process
        "general": "#95a5a6",  # Gray - miscellaneous
    }

    # Position clusters in 3D space with meaningful separation
    cluster_centers = {}
    total_categories = len(categorized_nodes)

    for i, category in enumerate(categorized_nodes.keys()):
        # Arrange clusters in a sphere with good visual separation
        if total_categories == 1:
            theta, phi = 0, 0
        else:
            theta = (2 * math.pi * i) / total_categories  # Horizontal distribution
            phi = math.pi * (0.3 + 0.4 * (i % 3) / 2)  # Vertical layers

        cluster_radius = 15  # Distance from origin for cluster centers
        center_x = cluster_radius * math.sin(phi) * math.cos(theta)
        center_y = cluster_radius * math.sin(phi) * math.sin(theta)
        center_z = cluster_radius * math.cos(phi)

        cluster_centers[category] = (center_x, center_y, center_z)

    # Generate nodes within their semantic clusters
    for category, items in categorized_nodes.items():
        center_x, center_y, center_z = cluster_centers[category]
        cluster_spread = min(8, max(3, len(items) * 0.5))  # Adaptive cluster size

        for i, item in enumerate(items):
            # Position nodes within cluster with slight randomization
            local_theta = random.uniform(0, 2 * math.pi)
            local_phi = random.uniform(0, math.pi)
            local_radius = random.uniform(1, cluster_spread)

            # Calculate final position relative to cluster center
            local_x = local_radius * math.sin(local_phi) * math.cos(local_theta)
            local_y = local_radius * math.sin(local_phi) * math.sin(local_theta)
            local_z = local_radius * math.cos(local_phi)

            final_x = center_x + local_x
            final_y = center_y + local_y
            final_z = center_z + local_z

            # Calculate importance based on content richness
            content_length = len(item["content"])
            title_length = len(item["title"])
            importance = min(1.0, (content_length + title_length * 2) / 300 + 0.2)

            nodes_data.append(
                {
                    "id": item["id"],
                    "title": item["title"],
                    "category": category,
                    "content": item["content"],
                    "source_table": item["source_table"],
                    "x": final_x,
                    "y": final_y,
                    "z": final_z,
                    "connections": 0,  # Will be calculated later
                    "importance": importance,
                    "color": category_colors.get(category, "#95a5a6"),
                    "cluster_center": cluster_centers[category],
                }
            )

    return nodes_data


def _calculate_semantic_connections(nodes_data: List[Dict]) -> List[Dict]:
    """
    Calculate meaningful connections between nodes based on semantic similarity
    and content relationships rather than random connections.
    """

    edges_data = []

    # Calculate connections based on multiple factors
    for i, node_a in enumerate(nodes_data):
        connections_added = 0
        max_connections = min(4, len(nodes_data) // 10)  # Adaptive connection limit

        for j, node_b in enumerate(nodes_data[i + 1:], i + 1):
            if connections_added >= max_connections:
                break

            connection_strength = 0.0
            connection_type = ""

            # Very strong connection: Same semantic category
            if node_a["category"] == node_b["category"]:
                connection_strength = 0.9
                connection_type = "semantic_cluster"

            # Strong connection: Related categories
            elif _are_categories_related(node_a["category"], node_b["category"]):
                connection_strength = 0.7
                connection_type = "related_domain"

            # Medium connection: Same source table (temporal/structural relationship)
            elif node_a["source_table"] == node_b["source_table"]:
                connection_strength = 0.5
                connection_type = "structural"

            # Weak connection: Content keywords overlap
            elif _content_similarity(node_a["content"], node_b["content"]) > 0.3:
                connection_strength = 0.4
                connection_type = "content_similar"

            # Create edge if connection is strong enough
            if connection_strength >= 0.4:
                edges_data.append(
                    {
                        "source": node_a["id"],
                        "target": node_b["id"],
                        "strength": connection_strength,
                        "type": connection_type,
                        "color": _get_edge_color(connection_type),
                    }
                )
                connections_added += 1

                # Update connection counts
                node_a["connections"] = node_a.get("connections", 0) + 1
                node_b["connections"] = node_b.get("connections", 0) + 1

    return edges_data


def _are_categories_related(cat1: str, cat2: str) -> bool:
    """Check if two semantic categories are conceptually related"""
    related_groups = [
        {"architecture", "api_design", "database"},  # System design
        {"deployment", "configuration", "workflow"},  # Operations
        {"testing", "error_handling", "performance"},  # Quality assurance
        {"development", "documentation", "workflow"},  # Development process
        {"security", "api_design", "architecture"},  # Security design
    ]

    for group in related_groups:
        if cat1 in group and cat2 in group:
            return True
    return False


def _content_similarity(content1: str, content2: str) -> float:
    """Calculate basic content similarity based on common words"""
    words1 = set(content1.lower().split())
    words2 = set(content2.lower().split())

    if not words1 or not words2:
        return 0.0

    intersection = words1.intersection(words2)
    union = words1.union(words2)

    return len(intersection) / len(union) if union else 0.0


def _get_edge_color(connection_type: str) -> str:
    """Get color for edge based on connection type"""
    colors = {
        "semantic_cluster": "#e74c3c",  # Red - strongest connections
        "related_domain": "#f39c12",  # Orange - related concepts
        "structural": "#3498db",  # Blue - structural relationships
        "content_similar": "#95a5a6",  # Gray - weak content similarity
    }
    return colors.get(connection_type, "#95a5a6")


def _has_content_similarity(content1: str, content2: str) -> bool:
    """Check if two content strings have basic similarity based on common keywords."""
    if not content1 or not content2:
        return False

    # Simple keyword-based similarity check
    words1 = set(content1.lower().split())
    words2 = set(content2.lower().split())

    # Remove common words that don't indicate meaningful similarity
    common_words = {
        "the",
        "a",
        "an",
        "and",
        "or",
        "but",
        "in",
        "on",
        "at",
        "to",
        "for",
        "of",
        "with",
        "by",
        "is",
        "are",
        "was",
        "were",
        "be",
        "been",
        "have",
        "has",
        "had",
        "do",
        "does",
        "did",
        "will",
        "would",
        "could",
        "should",
        "may",
        "might",
        "can",
        "this",
        "that",
        "these",
        "those",
    }
    words1 = words1 - common_words
    words2 = words2 - common_words

    if len(words1) < 3 or len(words2) < 3:
        return False

    # Calculate Jaccard similarity
    intersection = len(words1.intersection(words2))
    union = len(words1.union(words2))

    return (intersection / union) > 0.15 if union > 0 else False


# Implementation aliases for internal use
_create_interactive_d3_graph_impl = create_interactive_d3_graph
_create_advanced_d3_dashboard_impl = create_advanced_d3_dashboard
_export_graph_data_impl = export_graph_data
